/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    NaiveBayesSimple.java
 *    Copyright (C) 1999 University of Waikato, Hamilton, New Zealand
 *
 */

package weka.classifiers.bayes;

import weka.classifiers.Classifier;
import weka.core.Attribute;
import weka.core.AttributeStats;
import weka.core.Capabilities;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.RevisionUtils;
import weka.core.TechnicalInformation;
import weka.core.TechnicalInformationHandler;
import weka.core.Utils;
import weka.core.Capabilities.Capability;
import weka.core.TechnicalInformation.Field;
import weka.core.TechnicalInformation.Type;

import java.util.*;
import java.io.*;

/**
 <!-- globalinfo-start -->
 * Class for building and using a simple Naive Bayes classifier.Numeric attributes are modelled by a normal distribution.<br/>
 * <br/>
 * For more information, see<br/>
 * <br/>
 * Richard Duda, Peter Hart (1973). Pattern Classification and Scene Analysis. Wiley, New York.
 * <p/>
 <!-- globalinfo-end -->
 *
 <!-- technical-bibtex-start -->
 * BibTeX:
 * <pre>
 * &#64;book{Duda1973,
 *    address = {New York},
 *    author = {Richard Duda and Peter Hart},
 *    publisher = {Wiley},
 *    title = {Pattern Classification and Scene Analysis},
 *    year = {1973}
 * }
 * </pre>
 * <p/>
 <!-- technical-bibtex-end -->
 *
 <!-- options-start -->
 * Valid options are: <p/>
 * 
 * <pre> -D
 *  If set, classifier is run in debug mode and
 *  may output additional info to the console</pre>
 * 
 <!-- options-end -->
 *
 * @author Eibe Frank (eibe@cs.waikato.ac.nz)
 * @version $Revision: 1.20 $ 
*/
public class CostSensitiveNaiveBayes 
  extends Classifier
  implements TechnicalInformationHandler {
  
  /** for serialization */
  static final long serialVersionUID = -1478242251770381214L;

  /** All the counts for nominal attributes. */
  protected double [][][] m_Counts;
  
  /** Count of no. of occurences of a value j for attribute i */
  protected double [][] m_Frequency;
  
  /** Distinct values of categorical attribute */
  protected Set<Double> []m_DistinctValues;
  
  
  /** Max value of numeric attribute */
  protected double []m_MaxValue;
  
  /** Min value of numeric attribute */
  protected double []m_MinValue;
  
  /** Step for numeric attribute */
  protected double [] m_StepValue;  
  
  /** The mean for numeric attribute. */
  protected double [] m_Mean;

  /** The standard deviation for numeric attribute. */
  protected double [] m_Dev;

  
  /** The means for numeric attributes. */
  protected double [][] m_Means;

  /** The standard deviations for numeric attributes. */
  protected double [][] m_Devs;

  /** The prior probabilities of the classes. */
  protected double [] m_Priors;
  
  /** The number of occurences of each class */ 
  protected double [] m_ClassCount;  

  /** The instances used for training. */
  protected Instances m_Instances;

  /** Constant for normal distribution. */
  protected static double NORM_CONST = Math.sqrt(2 * Math.PI);
  
  /** Misclassification costs */  
  protected double C[][];
  
  /** Attribute costs */
  protected double T[];  
  

  /**
   * Returns a string describing this classifier
   * @return a description of the classifier suitable for
   * displaying in the explorer/experimenter gui
   */
  public String globalInfo() {
    return 
        "Class for building and using a simple Naive Bayes classifier."
      + "Numeric attributes are modelled by a normal distribution.\n\n"
      + "For more information, see\n\n"
      + getTechnicalInformation().toString();
  }

  /**
   * Returns an instance of a TechnicalInformation object, containing 
   * detailed information about the technical background of this class,
   * e.g., paper reference or book this class is based on.
   * 
   * @return the technical information about this class
   */
  public TechnicalInformation getTechnicalInformation() {
    TechnicalInformation 	result;
    
    result = new TechnicalInformation(Type.BOOK);
    result.setValue(Field.AUTHOR, "Richard Duda and Peter Hart");
    result.setValue(Field.YEAR, "1973");
    result.setValue(Field.TITLE, "Pattern Classification and Scene Analysis");
    result.setValue(Field.PUBLISHER, "Wiley");
    result.setValue(Field.ADDRESS, "New York");
    
    return result;
  }

  /**
   * Returns default capabilities of the classifier.
   *
   * @return      the capabilities of this classifier
   */
  public Capabilities getCapabilities() {
    Capabilities result = super.getCapabilities();

    // attributes
    result.enable(Capability.NOMINAL_ATTRIBUTES);
    result.enable(Capability.NUMERIC_ATTRIBUTES);
    //result.enable(Capability.DATE_ATTRIBUTES);
    result.enable(Capability.MISSING_VALUES);

    // class
    result.enable(Capability.NOMINAL_CLASS);
    result.enable(Capability.MISSING_CLASS_VALUES);
    
    return result;
  }
  
  
 
  
  
  
  public void setCosts()
  {
	  //C=new double[2][2];
	  //T=new double[9];
	  int numClasses=m_Instances.numClasses();
	  C=new double[numClasses][numClasses];
	  T=new double[m_Instances.numAttributes()-1];
	  String attrFile;
	  String costFile;
	  double costAttrib[]=T;
	  double costMatrix[][]=C;
	
	  try
		{
			//BufferedReader bf;			
			BufferedReader bf=new BufferedReader(new FileReader("costs.cfg"));
			attrFile=bf.readLine();
			costFile=bf.readLine();
			bf.close();
//			attrFile="/users1/pg08/pkamle/weka/wekaProj/data/csdataset.attr.cost";
//			costFile="/users1/pg08/pkamle/weka/wekaProj/data/csdataset.cost";
			
			
			int i=0;
			//read attribute costs
			bf=new BufferedReader(new FileReader(attrFile));
			while(bf.ready())
			{
				String str[]=bf.readLine().split("\\s+");
				costAttrib[i]=Double.parseDouble(str[1]);
				//System.out.println("Attrib Cost: "+costAttrib[i]);
				i++;
			}
			bf.close();

			//read misclassification costs
			bf=new BufferedReader(new FileReader(costFile));
			//Ignore firest 3 lines
			bf.readLine();
			bf.readLine();
			bf.readLine();

			i=0;
			while(bf.ready() && i<numClasses)
			{
				String str[]=bf.readLine().split("\\s+");
				for(int j=0;j<numClasses;j++)
				{
					costMatrix[i][j]=Double.parseDouble(str[j]);
					//System.out.println("Cost["+i+","+j+"]="+costMatrix[i][j]);
				}
				i++;           
			}
			bf.close();

		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
	}
  
 
  
  
  protected boolean isEmpty(boolean arr[])
  {
	  for(int i=0;i<arr.length;i++)
	  {
		  if(arr[i]==true)
			  return false;  //not empty
	  }
	  	  
	  return true;
  }
  
  
  /** probability that attribute A_attIndex = attValue give class classIndex */
  protected double pAttributeGivenClass(int attIndex, double attValue, int classIndex)
  {
	  if(m_Instances.attribute(attIndex).isNumeric())
		  return normalDens(attValue,m_Means[classIndex][attIndex], m_Devs[classIndex][attIndex]);
	  else
	  {
		  return m_Counts[classIndex][attIndex][(int)attValue]/m_ClassCount[classIndex];
	  }
  }
  
  protected double pAttributeGivenClass(boolean attList[], Instance instance, int classIndex)
  {
	  //OK
	  double prob=1;
//	  int working=0;
	  
	  for(int i=0;i<attList.length; i++)
		  if(attList[i] && (!instance.isMissing(i)))
		  {
//			  working=1;
			  prob *= pAttributeGivenClass(i,instance.value(i) ,classIndex);
		  }
	  
//	  if (prob*working!=1){
//		  System.out.println("Attrib_class: "+prob*working);
//	  }
	  //return prob*working;
	  return prob;
  }
  
  protected double pClass(int i)
  {
	  //OK
	  double temp=(double)m_ClassCount[i]/(double)m_Instances.numInstances();
	  //System.out.println("pClass: "+temp);
	  return temp;
  }
  
  protected double pAttribute(int attIndex, double value)
  {
	  //OK
	  //System.out.println("Here!");
	  if(m_Instances.attribute(attIndex).isNumeric())
	  {
		//  System.out.println("pAttr:"+normalDens(value, m_Mean[attIndex],m_Dev[attIndex]));
		  return normalDens(value, m_Mean[attIndex],m_Dev[attIndex]);
	  }
	  else
		  return m_Frequency[attIndex][(int)value]/(double)(m_Instances.numInstances());
	  
  }
  
  protected double pAttribute(boolean attList[], Instance instance)
  {
	  //OK
	  double prob=1;
	  for(int i=0;i<attList.length; i++)
		  if(attList[i] && (!instance.isMissing(i)))
		  {
			  prob *= pAttribute(i,instance.value(i));
		  }
	  //System.out.println("pAttribute: "+prob);
	  return prob;
  }
  
  /** P(C_j|A_tilde) = P(C_j).P(A_tilde|C_j)/P(A_tilde) */
  protected double pClassGivenAttribute(int classIndex, boolean aTilde[], Instance instance) 
  {
	  //PK
	  //P(C_j|A_tilde) = P(C_j).P(A_tilde|C_j)/P(A_tilde)
  
	  //System.out.println("pcga: "+ pClass(classIndex) + " * " + pAttributeGivenClass(aTilde, instance, classIndex) + " / " + pAttribute(aTilde, instance));
	  double temp=pClass(classIndex)*pAttributeGivenClass(aTilde, instance, classIndex) /pAttribute(aTilde, instance);

	  //return (temp<1?temp:1);
	  return temp;
	  
	  //return pClass(classIndex)*pAttributeGivenClass(aTilde, instance, classIndex)/pAttribute(aTilde, instance);
	  
  }
  
  /** Summation_i_class (C_ij.P(C_i|A_tilde)) */
  protected double riskOfClassGivenAttributes(int classIndex,Instance instance, boolean[] aTilde)
  {
	  double risk=0;
	  for(int i=0;i<m_Instances.numClasses();i++){	  
		  //System.out.println("Risko: "+ C[i][classIndex] + " * " + pClassGivenAttribute(classIndex, aTilde, instance));		  
		  risk+=C[i][classIndex]*pClassGivenAttribute(i, aTilde, instance);
	  }
	  return risk;
  }
  
  protected double[] getRisksForInstance(Instance instance, boolean[] aTilde)
  {   
	  double[] risks=new double[m_Instances.numClasses()]; 
	  for(int i=0;i<m_Instances.numClasses();i++)
	  {
		  risks[i]=riskOfClassGivenAttributes(i, instance, aTilde);
	  }
	  
	   // Normalize risks
	   //Utils.normalize(risks);	    
	  return risks;  
  }
  

  protected double[] getDistinctValues(int attIndex)
  {
	  
	  int n=m_DistinctValues[attIndex].size();
	  
	  double[] vals=new double[n]; 
	  Iterator<Double> iter= m_DistinctValues[attIndex].iterator();
	  
	  int i=0;
	  while(iter.hasNext())
	  {
		  vals[i]=iter.next();
		  i++;		  
	  }
	  return vals;	  
	  
  }
  
   
  
  protected double getMisclassificationCost(Instance instance, boolean attList[], int attIndex)
  {
	  double cost=0;
	  
	  if(m_Instances.attribute(attIndex).isNumeric())
	  {
//		  Enumeration e=m_Instances.enumerateInstances();
//		  
//		  attList[attIndex]=true;
//		  double val;
//		  while(e.hasMoreElements())
//		  {
//			  Instance v=(Instance)e.nextElement();
//			  val=v.value(attIndex);
//			  double[] vals=instance.toDoubleArray();
//			  double wt=instance.weight();
//			  vals[attIndex]=val;
//			  
//			  Instance temp=new Instance(wt,vals);		  
//			  cost+=pAttribute(attIndex, val)*getMisclassificationCost(temp, attList);
//			  
//		  }
//		  attList[attIndex]=false;
//		  
		  attList[attIndex]=true;
		  for(double val=m_MinValue[attIndex]; val<m_MaxValue[attIndex]; val+=m_StepValue[attIndex])
		  {
			  double[] vals=instance.toDoubleArray();
			  double wt=instance.weight();
			  vals[attIndex]=val;
			  
			  Instance temp=new Instance(wt,vals);		  
			  cost+=pAttribute(attIndex, val)*getMisclassificationCost(temp, attList);
			  
		  }
		  attList[attIndex]=false;
	  }
	  else
	  {
	  
		  double distinctValues[]=getDistinctValues(attIndex);
		  
		  attList[attIndex]=true;
		  for(int j=0;j<distinctValues.length;j++)
		  {
			  double[] vals=instance.toDoubleArray();
			  double wt=instance.weight();
			  vals[attIndex]=distinctValues[j];
			  
			  Instance temp=new Instance(wt,vals);		  
			  cost+=pAttribute(attIndex, distinctValues[j])*getMisclassificationCost(temp, attList);
			  
		  }
		  attList[attIndex]=false;
	  }
	  return cost;  
  }
  
  protected double getMisclassificationCost(Instance instance,boolean attList[])
  {
	  double misclassificationCosts[]= getRisksForInstance(instance, attList);
	  
	  double min=Double.MAX_VALUE;
	  //double min=0;
	  for(int i=0;i<misclassificationCosts.length;i++)	
	  {
		  if(misclassificationCosts[i]<min)
		  {
			  //System.out.println("Miscost: " + misclassificationCosts[i] + " * " + m_Priors[i]);
			
			  //This code gives average misclassification
			  //min+=misclassificationCosts[i]*m_Priors[i];
			
			  min=misclassificationCosts[i];
		  }
	  }
	  return min;	  
	  
  }
  
  
  protected double gain(Instance instance,boolean aTilde[],int i)
  {
	  double x1= getMisclassificationCost(instance, aTilde);
	  
	  double x2= getMisclassificationCost(instance, aTilde,i);	  
	  
	  //System.out.println(x1 + " - (" + x2 + ") = "+(x1-x2));
	  
	  return x1-x2;
	  //System.out.println("Gain: " + (x2-x1));
	  //return (x2-x1);
  }
    
  protected double getUtilityOfAttribute(int i,boolean aTilde[], Instance instance)
  {
	  return gain(instance,aTilde,i)-T[i];
  }
  
  protected boolean thereExistsGreaterThanZero(double arr[], boolean []selected)
  {	  	  
	  for(int i=0;i<arr.length;i++)
		  if(!selected[i])
		  {
			  if(arr[i]>0)
				  return true;
		  }
	  
	  return false;
  }
  
  protected int max(double arr[], boolean atilde[])
  {
	  double max=-Double.MAX_VALUE;
	  int maxIndex=-1;
	  for(int i=0;i<arr.length;i++)
	  {
		  if (atilde[i] || arr[i]<=max)
			  continue;
		  
//		  if(arr[i]>max)
//		  {
			  max=arr[i];
			  maxIndex=i;
//		  }
	  }
	  
	  return maxIndex;
  }
  
  
  
  public double classifyInstance(Instance instance) throws Exception
  {
	  //setCosts();
	  boolean aTilde[]=new boolean[instance.numAttributes()-1];
	  boolean aBar[]=new boolean[instance.numAttributes()-1];
	  
	  //Initialize aTilde and aBar
	  for(int i=0;i<instance.numAttributes()-1;i++)
	  {
		  aTilde[i]=false;
		  aBar[i]=true;
	  }
	  
	  double c_test=0;
	  
	  while(!isEmpty(aBar))
	  {
		  double util_aBar[] = new double[aBar.length];
		  for(int i=0;i<aBar.length; i++)
			  if(aBar[i]==true)
			  {
				  //System.out.println(instance.value(i));
				  util_aBar[i] = getUtilityOfAttribute(i,aTilde,instance);
			  }
			  else
			  {
				  util_aBar[i]=-1;
			  }
		  //for(int m=0;m<util_aBar.length;m++)
			 // System.out.print("Utilities:"+ m +" "+util_aBar[m]);
		  
		  if(!thereExistsGreaterThanZero(util_aBar,aTilde))  
			  break;
		  
	
		  int aistar=max(util_aBar,aTilde);
		  //System.out.println(" Selected attr: "+aistar);
		  //v = instance.attribute(aistar).
		  c_test+=T[aistar];
		  
		  aTilde[aistar]=true;
		  aBar[aistar]=false;  		  
	  }
	  
	  double misclassificationCosts[]=getRisksForInstance(instance, aTilde);

	  double min=Double.MAX_VALUE;
	  int minIndex=-1;
	  for(int i=0;i<misclassificationCosts.length;i++)	
	  {
		  
		  //System.out.println("Misclassify[" + i + "] = " + misclassificationCosts[i]);
		  
		  if(misclassificationCosts[i]<min)
		  {
			  min=misclassificationCosts[i];
			  minIndex=i;
		  }
	  }	  	  
	  //System.out.println("MinIndex: " + minIndex);
	  return minIndex;
  }
  
  
  /**
   * Generates the classifier.
   *
   * @param instances set of instances serving as training data 
   * @exception Exception if the classifier has not been generated successfully
   */
  public void buildClassifier(Instances instances) throws Exception {

    int attIndex = 0;
    double sum;

    // can classifier handle the data?
    getCapabilities().testWithFail(instances);

    // remove instances with missing class
    instances = new Instances(instances);
    instances.deleteWithMissingClass();

    m_Instances = new Instances(instances, instances.numInstances());
    Enumeration<Instance> enumer =  instances.enumerateInstances();
    while(enumer.hasMoreElements())
    	m_Instances.add(enumer.nextElement());
    
    setCosts(); //my line
    
    
    //System.out.print(m_Instances.numInstances());
    
    // Reserve space
    
    //my lines begin
    m_ClassCount = new double[instances.numClasses()]; //needs deletion
    
    m_DistinctValues =  new TreeSet[instances.numAttributes()-1]; 
    
    for (int i = 0; i<m_DistinctValues.length; i++) 
    {
    	m_DistinctValues[i] = new TreeSet<Double> ();
    }
    
    m_Frequency = new double[instances.numAttributes()-1][];
    
    m_MaxValue = new double[instances.numAttributes()-1];
    m_MinValue = new double[instances.numAttributes()-1];
    m_StepValue = new double[instances.numAttributes()-1];
    m_Mean = new double[instances.numAttributes()-1];
    m_Dev =new double[instances.numAttributes()-1];
    
    //my lines end
    
    m_Counts = new double[instances.numClasses()]
      [instances.numAttributes() - 1][0];
    m_Means = new double[instances.numClasses()][instances.numAttributes() - 1];
    m_Devs = new double[instances.numClasses()][instances.numAttributes() - 1];
    m_Priors = new double[instances.numClasses()];
    
    Enumeration enu = instances.enumerateAttributes();
    while (enu.hasMoreElements()) 
    {
    	Attribute attribute = (Attribute) enu.nextElement();
    	if (attribute.isNominal()) 
    	{
    		for (int j = 0; j < instances.numClasses(); j++) 
    		{
    			m_Counts[j][attIndex] = new double[attribute.numValues()];
    		}
    		m_Frequency[attIndex] = new double[attribute.numValues()]; // my line
    	} 
    	else 
    	{
    		for (int j = 0; j < instances.numClasses(); j++) 
    		{
    			m_Counts[j][attIndex] = new double[1];
    		}
    	}
    	attIndex++;
    }
    
    // Compute counts and sums
    Enumeration enumInsts = instances.enumerateInstances();
    while (enumInsts.hasMoreElements()) 
    {
    	Instance instance = (Instance) enumInsts.nextElement();
    	if (!instance.classIsMissing()) 
    	{
    		Enumeration enumAtts = instances.enumerateAttributes();
    		attIndex = 0;
    		while (enumAtts.hasMoreElements()) 
    		{
    			Attribute attribute = (Attribute) enumAtts.nextElement();
    			if (!instance.isMissing(attribute)) 
    			{
    				if (attribute.isNominal()) 
    				{
    					m_Counts[(int)instance.classValue()][attIndex]
    					                                     [(int)instance.value(attribute)]++;
    					m_Frequency[attIndex][(int)instance.value(attribute)]++; //my line
    					m_DistinctValues[attIndex].add(instance.value(attribute)); // my line
    				}
    				else 
    				{
    					m_Means[(int)instance.classValue()][attIndex] +=
    						instance.value(attribute);
    					m_Counts[(int)instance.classValue()][attIndex][0]++;
    				}
    			}
    			attIndex++;
    		}
    		m_Priors[(int)instance.classValue()]++;
    		m_ClassCount[(int)instance.classValue()]++;  //deletion reqd
    	}
    }
    
    // Compute means
    Enumeration enumAtts = instances.enumerateAttributes();
    attIndex = 0;
    while (enumAtts.hasMoreElements()) 
    {
    	Attribute attribute = (Attribute) enumAtts.nextElement();
    	if (attribute.isNumeric()) 
    	{
    		m_Mean[attIndex]=instances.meanOrMode(attIndex);  //my line
    		m_Dev[attIndex]=Math.sqrt(instances.variance(attIndex)); //my line
    		
    		m_MaxValue[attIndex]=instances.attributeStats(attIndex).numericStats.max;  //my line
    		m_MinValue[attIndex]=instances.attributeStats(attIndex).numericStats.min;  //my line
    		m_StepValue[attIndex]=0.1*(m_MaxValue[attIndex]-m_MinValue[attIndex]);  //my line
    		
    		for (int j = 0; j < instances.numClasses(); j++) 
    		{
    			if (m_Counts[j][attIndex][0] < 2) 
    			{
    				throw new Exception("attribute " + attribute.name() +
    						": less than two values for class " +
    						instances.classAttribute().value(j));
    			}
    			m_Means[j][attIndex] /= m_Counts[j][attIndex][0];
    		}
    	}
    	attIndex++;
    }    
    
    // Compute standard deviations
    enumInsts = instances.enumerateInstances();
    while (enumInsts.hasMoreElements()) 
    {
    	Instance instance =	(Instance) enumInsts.nextElement();
    	if (!instance.classIsMissing()) 
    	{
    		enumAtts = instances.enumerateAttributes();
    		attIndex = 0;
    		while (enumAtts.hasMoreElements()) 
    		{
    			Attribute attribute = (Attribute) enumAtts.nextElement();
    			if (!instance.isMissing(attribute)) 
    			{
    				if (attribute.isNumeric()) 
    				{
    					m_Devs[(int)instance.classValue()][attIndex] += (m_Means[(int)instance.classValue()][attIndex]-
    								instance.value(attribute))*(m_Means[(int)instance.classValue()][attIndex]-
    										instance.value(attribute));
    				}
    			}
    			attIndex++;
    		}
    	}
    }
    enumAtts = instances.enumerateAttributes();
    attIndex = 0;
    while (enumAtts.hasMoreElements()) {
    	Attribute attribute = (Attribute) enumAtts.nextElement();
    	if (attribute.isNumeric()) {
    		for (int j = 0; j < instances.numClasses(); j++) {
    			if (m_Devs[j][attIndex] <= 0) {
    				throw new Exception("attribute " + attribute.name() +
    						": standard deviation is 0 for class " +
    						instances.classAttribute().value(j));
    			}
    			else {
    				m_Devs[j][attIndex] /= m_Counts[j][attIndex][0] - 1;
    				m_Devs[j][attIndex] = Math.sqrt(m_Devs[j][attIndex]);
    			}
    		}
    	}
    	attIndex++;
    } 

    // Normalize counts
    enumAtts = instances.enumerateAttributes();
    attIndex = 0;
    while (enumAtts.hasMoreElements())
    {
    	Attribute attribute = (Attribute) enumAtts.nextElement();
    	if (attribute.isNominal()) 
    	{
    		for (int j = 0; j < instances.numClasses(); j++) 
    		{
    			sum = Utils.sum(m_Counts[j][attIndex]);
    			for (int i = 0; i < attribute.numValues(); i++) 
    			{
    				m_Counts[j][attIndex][i] =(m_Counts[j][attIndex][i] + 1)/(sum + (double)attribute.numValues());
    				
    			}
    		}
    	}
    	attIndex++;
    }
    
    // Normalize priors
    sum = Utils.sum(m_Priors);
    for (int j = 0; j < instances.numClasses(); j++)
      m_Priors[j] = (m_Priors[j] + 1)/(sum + (double)instances.numClasses());

  }



  /**
   * Returns a description of the classifier.
   *
   * @return a description of the classifier as a string.
   */
  public String toString() {

    if (m_Instances == null) {
      return "Naive Bayes (simple): No model built yet.";
    }
    try {
      StringBuffer text = new StringBuffer("Naive Bayes (simple)");
      int attIndex;
      
      for (int i = 0; i < m_Instances.numClasses(); i++) {
	text.append("\n\nClass " + m_Instances.classAttribute().value(i) 
		    + ": P(C) = " 
		    + Utils.doubleToString(m_Priors[i], 10, 8)
		    + "\n\n");
	Enumeration enumAtts = m_Instances.enumerateAttributes();
	attIndex = 0;
	while (enumAtts.hasMoreElements()) {
	  Attribute attribute = (Attribute) enumAtts.nextElement();
	  text.append("Attribute " + attribute.name() + "\n");
	  if (attribute.isNominal()) {
	    for (int j = 0; j < attribute.numValues(); j++) {
	      text.append(attribute.value(j) + "\t");
	    }
	    text.append("\n");
	    for (int j = 0; j < attribute.numValues(); j++)
	      text.append(Utils.
			  doubleToString(m_Counts[i][attIndex][j], 10, 8)
			  + "\t");
	  } else {
	    text.append("Mean: " + Utils.
			doubleToString(m_Means[i][attIndex], 10, 8) + "\t");
	    text.append("Standard Deviation: " 
			+ Utils.doubleToString(m_Devs[i][attIndex], 10, 8));
	  }
	  text.append("\n\n");
	  attIndex++;
	}
      }
      
      return text.toString();
    } catch (Exception e) {
      return "Can't print Naive Bayes classifier!";
    }
  }

  /**
   * Density function of normal distribution.
   * 
   * @param x the value to get the density for
   * @param mean the mean
   * @param stdDev the standard deviation
   * @return the density
   */
  protected double normalDens(double x, double mean, double stdDev) {
    
    double diff = x - mean;
    
    return (1 / (NORM_CONST * stdDev)) 
      * Math.exp(-(diff * diff / (2 * stdDev * stdDev)));
  }
  
  /**
   * Returns the revision string.
   * 
   * @return		the revision
   */
  public String getRevision() {
    return RevisionUtils.extract("$Revision: 1.20 $");
  }

  /**
   * Main method for testing this class.
   *
   * @param argv the options
   */
  public static void main(String [] argv) {
    runClassifier(new CostSensitiveNaiveBayes(), argv);
  }
}
