/*Borole Classifier 001*/

package weka.classifiers.functions;

import weka.classifiers.Classifier;
import weka.core.Capabilities;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.RevisionUtils;
import weka.core.WeightedInstancesHandler;
import weka.core.Capabilities.Capability;

import Jama.Matrix;

import riso.numerical.LBFGS;

/**
This is the basic abstract class that uses LBFGS for optimisation, and
solves the linear regression problem. It needs 2 functions to be implemented.
The first is the value of the loss function: double getErrorFunction()
and the second is the gradient of the loss function: Matrix getGradient();
 */
public abstract class CustomLinearRegression extends Classifier implements WeightedInstancesHandler {

	private static final long serialVersionUID = -6632048542000555464L;
	
/////////////////INTERNAL VARIABLES/////////////////  
  protected int ndim;//number of attributes (including the class)
  protected int n;//number of samples
  protected Matrix x;//the attributes for all samples (n x ndim matrix, with first element=1 for each sample)
  protected Matrix y;//the class value for all samples (n x 1 matrix) 
  protected Matrix w;//the resultant weights (ndim x 1 matrix)
  
  Instances attributeNames; //for getting the names of the attribs from the arff file.
  
  
//////////////////////////////////////////////////

  private void readData(Instances insts){
	  	ndim = insts.numAttributes();
	  	n = insts.numInstances();
	  	
		x = new Matrix(n, ndim);
		y = new Matrix(n, 1);
	
		// read data set
		for (int i = 0; i < n; i++)//go through each instance of data
		{
			x.set(i,0,1);
			for (int j=1;j<ndim;j++){
				x.set(i, j,  insts.instance(i).value(j-1));
			}
			y.set(i, 0, insts.instance(i).value(ndim-1));
		}
		
		attributeNames = new Instances(insts, 0);
  }

  protected abstract double getErrorFunction();
  
  protected abstract Matrix getGradient();

  private void doRegression()
  {
	  int j;
	  Matrix gradient, diag;
	  //w = Matrix.random(ndim, 1).times(5);
	  w = new Matrix(ndim, 1);
	
	  diag = new Matrix(1, ndim);
	
	  double f;
	  double eps, xtol;
	  int[] iprint, iflag = new int[1];
	  
	  int icall;
	  int numCorrections;
	  iprint = new int[2];
	  boolean diagco;
	
	  numCorrections = 100;
	  iprint[0] = 0;
	  iprint[1] = 0;
	  diagco = false;
	  eps = 1.0e-1;
	  xtol = 1.0 / Double.MAX_VALUE;
	  icall = 0;
	  iflag[0] = 0;
	
	  /* initialize W. */
	
	  for (j = 0; j < ndim; j++){
		  w.set(j, 0, 0.00001);//ndim * new Random().nextDouble();
	  }
	
	  do
	  {
		  //f = 0;
		  /* find the value of error function for current w */			
		  f = getErrorFunction();
		  //	System.out.println("sqe = " + f);
	
		  /* find gradient of F(w) at current w
		   * 
		   *for square loss, F'(W) = X' * (XW - Y)
		   * */
		  gradient = getGradient();
	
		  try
		  {
			  double[][] w_parameter = w.transpose().getArray();
			  LBFGS.lbfgs(ndim, numCorrections, w_parameter[0], f, gradient.transpose().getArray()[0], diagco, diag.getArray()[0], iprint, eps, xtol, iflag);
			  w = new Matrix(w_parameter).transpose();
			  
		  }
		  catch (LBFGS.ExceptionWithIflag e)
		  {
			  System.err.println("lbfgs done :)\n" + e);
			  return;
		  }
		  icall += 1;
		    
	  }
	  while (iflag[0] != 0/* && icall <= 4000 */);
	  	
  }  
  
  /**
   * Returns a string describing this classifier
   * @return a description of the classifier suitable for
   * displaying in the explorer/experimenter gui
   */
  public String globalInfo() {
    return "This is a Generic Linear Regression module implemented by" +
    "Anup Kulkarni, Prashant Borole, Prashant Kamle and Sriram Kashyap" +
    "as part of a Data Mining Assignment. To use it, extend the class" +
    "and implement the error function, and the gradient function";
  }

  /**
   * Generate a prediction for the supplied instance.
   *
   * @param inst the instance to predict.
   * @return the prediction
   * @throws Exception if an error occurs
   */
  public double classifyInstance(Instance inst) throws Exception {
	  Matrix temp1=new Matrix(ndim,1);
	  temp1.set(0, 0, 1);
	  for (int i=1;i<ndim;i++){
		  temp1.set(i, 0, inst.value(i-1));
	  }
    return w.transpose().times(temp1).get(0,0);
  }

  /**
   * Returns default capabilities of the classifier.
   *
   * @return      the capabilities of this classifier
   */
  public Capabilities getCapabilities() {
    Capabilities result = super.getCapabilities();

    // attributes
    result.enable(Capability.NUMERIC_ATTRIBUTES);

    // class
    result.enable(Capability.NUMERIC_CLASS);
    
    return result;
  }
  
  /**
   * Builds a simple linear regression model given the supplied training data.
   *
   * @param insts the training data.
   * @throws Exception if an error occurs
   */
  public void buildClassifier(Instances insts) throws Exception {
	  readData(insts);
	  doRegression();
  }

  /**
   * Returns a description of this classifier as a string
   *
   * @return a description of the classifier.
   */
  public String toString() {

    StringBuffer text = new StringBuffer();
    text.append("Result of regression:\n");

    for (int i=1;i<w.getRowDimension();i++){
    	text.append(w.get(i,0) + "\t * " + attributeNames.attribute(i-1).name() + " + \n");
    }
	
    text.append(w.get(0,0) + "\n");
	
    
//    Matrix result1=w.transpose().times(x.transpose());
//	for (int i=0;i<result1.getColumnDimension();i++)
//	{
//		text.append(result1.get(0, i) + "\n"); //System.out.print(result1.get(i, j) + "\t");
//	}
    
    return text.toString();
  }
  
  /**
   * Returns the revision string.
   * 
   * @return		the revision
   */
  public String getRevision() {
    return RevisionUtils.extract("$Revision: 1.00 $");
  }

}
